/*
 *  dbload by Davide Libenzi (GDBM Database Load)
 *  Copyright (C) 2008  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <gdbm.h>



static char *dbld_get(FILE *fin, int *rsize);
static int dbld_loadrec(FILE *fin, datum *dkey, datum *ddata);
static int dbld_load(char const *dbfile, FILE *fin);
static void dbld_usage(char const *prg);



static char *dbld_get(FILE *fin, int *rsize) {
	int asize, size, c;
	char *blk, *rblk;
	char buf[8];

	if ((blk = (char *) malloc(asize = 1024)) == NULL)
		return NULL;
	for (size = 0;;) {
		if ((c = fgetc(fin)) == EOF) {
			free(blk);
			return NULL;
		}
		if (size + 2 >= asize) {
			asize = 2 * asize + 1;
			if ((rblk = (char *) realloc(blk, asize)) == NULL) {
				free(blk);
				return NULL;
			}
			blk = rblk;
		}
		if (c == '\\') {
			if ((c = fgetc(fin)) == EOF) {
				free(blk);
				return NULL;
			}
			if (c == 'x') {
				buf[0] = (char) fgetc(fin);
				buf[1] = (char) fgetc(fin);
				buf[2] = 0;
				if (sscanf(buf, "%x", &c) != 1) {
					free(blk);
					return NULL;
				}
			}
			blk[size++] = (char) c;
		} else if (c == ';' || c == '\n') {
			break;
		} else if (c == '\r') {
			if ((c = fgetc(fin)) != '\n') {
				free(blk);
				return NULL;
			}
		} else {
			blk[size++] = (char) c;
		}
	}
	blk[size] = 0;
	*rsize = size;

	return blk;
}

static int dbld_loadrec(FILE *fin, datum *dkey, datum *ddata) {
	int size;
	char *data;

	if ((data = dbld_get(fin, &size)) == NULL)
		return -1;
	dkey->dptr = data;
	dkey->dsize = size;
	if ((data = dbld_get(fin, &size)) == NULL) {
		free(dkey->dptr);
		return -2;
	}
	ddata->dptr = data;
	ddata->dsize = size;

	return 0;
}

static int dbld_load(char const *dbfile, FILE *fin) {
	int line, res;
	GDBM_FILE gdbf;
	datum dkey, ddata;

    	if ((gdbf = gdbm_open((char *) dbfile, 0, GDBM_WRCREAT, 00664, NULL)) == NULL) {
		fprintf(stderr, "unable to open DB file: file='%s'\n", dbfile);
		return -1;
	}
	for (line = 0; (res = dbld_loadrec(fin, &dkey, &ddata)) == 0; line++) {
		if (gdbm_store(gdbf, dkey, ddata, GDBM_REPLACE) != 0) {
			fprintf(stderr, "unable to store to DB file: file='%s'\n", dbfile);
			free(dkey.dptr);
			free(ddata.dptr);
			gdbm_close(gdbf);
			return -1;
		}
		free(dkey.dptr);
		free(ddata.dptr);
	}
	gdbm_close(gdbf);
	if (res == -2) {
		fprintf(stderr, "record load error: file='%s' line=%d\n", dbfile, line + 1);
		return -1;
	}

	return 0;
}

static void dbld_usage(char const *prg) {

	fprintf(stderr,
		"use: %s --dbfile DBFIL [--input IFIL] [--help]\n", prg);
}

int main(int ac, char **av) {
	int i, res;
	char const *dbfile = NULL, *ifile = NULL;
	FILE *fin = stdin;

	for (i = 1; i < ac; i++) {
		if (!strcmp(av[i], "--dbfile")) {
			if (++i < ac)
				dbfile = av[i];
		} else if (!strcmp(av[i], "--input")) {
			if (++i < ac)
				ifile = av[i];
		} else {
			dbld_usage(av[0]);
			return 1;
		}
	}
	if (dbfile == NULL) {
		dbld_usage(av[0]);
		return 1;
	}
	if (ifile != NULL && (fin = fopen(ifile, "r")) == NULL) {
		perror(ifile);
		return 2;
	}
	res = dbld_load(dbfile, fin);
	if (fin != stdin)
		fclose(fin);

	return res == 0 ? 0: 3;
}

